---- Customized scoring

require("glon")

SCORE = {}
SCORE.Events = {}
--SCORE.Scores = {}

-- One might wonder why all the key names in the event tables are so annoyingly
-- short. Well, the serialisation module in gmod (glon) does not do any
-- compression. At all. This means the difference between all events having a
-- "time_added" key versus a "t" key is very significant for the amount of data
-- we need to send. It's a pain, but I'm not going to code my own compression,
-- so doing it manually is the only way.

-- One great way to reduce data sent turned out to be rounding the time floats.
-- We don't actually need to know about 10000ths of seconds after all.

-- From gmod wiki
function math.AdvRound( val, d )
   d = d or 0;
   return math.Round( val * (10 ^ d) ) / (10 ^ d);
end

function SCORE:AddEvent(entry)
   entry["t"] = math.AdvRound(CurTime(), 2)
   table.insert(self.Events, entry)
end

local function WepToEnum(wep)
   local cls = wep:GetClass()
   if cls == "weapon_zm_pistol" then
      return AMMO_PISTOL
   elseif cls == "weapon_zm_shotgun" then
      return AMMO_SHOTGUN
   elseif cls == "weapon_zm_rifle" then 
      return AMMO_RIFLE
   elseif cls == "weapon_zm_revolver" then
      return AMMO_DEAGLE
   elseif cls == "weapon_zm_mac10" then
      return AMMO_MAC10
   elseif clas == "weapon_zm_improvised" then
      return AMMO_CROWBAR
   end
end

local function CopyDmg(dmg)
   local inf = dmg:GetInflictor()
   local wep = nil
   if ValidEntity(inf) then
      if inf:IsWeapon() then
         wep = inf
      elseif inf:IsPlayer() then
         wep = inf:GetActiveWeapon()
         if not ValidEntity(wep) then
            wep = nil
         end
      end
   end

   -- t = type, a = amount, g = gun, h = headshot
   local d = {}
   d.t = (dmg:IsFallDamage() and DMG_FALL) or (dmg:IsBulletDamage() and DMG_BULLET) or (dmg:IsExplosionDamage() and DMG_BLAST)
   d.a = dmg:GetDamage()
   d.h = false

   if wep then
      d.g = WepToEnum(wep)

      -- override for melee damage
      if wep:GetClass() == "weapon_zm_improvised" then
         d.t = DMG_CLUB
      end
   else
      d.g = -1
   end

   return d
end

function SCORE:HandleKill( victim, attacker, dmginfo )

   local e = {
      id=EVENT_KILL, 
      att={ni="", uid=-1, tr=false}, 
      vic={ni=victim:Nick(), uid=victim:UniqueID(), tr=false}, 
      dmg=CopyDmg(dmginfo)};

   if attacker:IsValid() and attacker:IsPlayer() then
      e.att.ni = attacker:Nick()
      e.att.uid = attacker:UniqueID()
      e.att.tr = attacker.is_traitor
   end

   if victim:IsValid() and victim:IsPlayer() then
      Msg("Kill headshot val: " .. tostring(victim.was_headshot) .. "\n")
      e.dmg.h = victim.was_headshot
   end

   e.vic.tr = victim.is_traitor

   self:AddEvent(e)
end

function SCORE:HandleSpawn( ply )
   if ply:Team() == TEAM_TERROR then
      self:AddEvent({id=EVENT_SPAWN, ni=ply:Nick(), uid=ply:UniqueID()})
   end
end

function SCORE:HandleTraitorSelection()
   local traitors = {}
   for k, ply in pairs(player.GetAll()) do
      if ply.is_traitor then
         table.insert(traitors, ply:UniqueID())
      end
   end

   self:AddEvent({id=EVENT_SELECTED, traitor_ids=traitors})
end

function SCORE:ApplyEventLogScores()
   local scores = {}
   for k, ply in pairs(player.GetAll()) do
      scores[ply:UniqueID()] = {}
   end

   local ply = nil
   for id, s in pairs(ScoreEventLog(self.Events, scores)) do
      ply = player.GetByUniqueID(id)
      if ply then
         ply:AddFrags(KillsToPoints(s, ply.is_traitor))
      end
   end
end

function SCORE:RoundStateChange(newstate)
   self:AddEvent({id=EVENT_GAME, state=newstate})
end

function SCORE:RoundComplete(wintype)
   self:AddEvent({id=EVENT_FINISH, win=wintype})
end

function SCORE:Reset()
   self.Events = {}
end

local function CompressForStream(events)
   -- may want to filter out data later
   -- just serialize for now

--   local h_a, h_b, h_c = debug.gethook()
--   debug.sethook()
   local status, result = pcall(glon.encode, events)
--   debug.sethook(h_a, h_b, h_c)

   if not status then
      Msg("Round report event encoding returned an error: " .. result .. "\n")
      return false
   else
      return result
   end
end

function SCORE:StreamToClients()
--   datastream.StreamToClients(player.GetHumans(), "ReportStreamHook", self.Events)
   -- datastream is a piece of crap, do it ourselves

   local s = CompressForStream(self.Events)

   if not s then
      return -- error occurred
   end

   -- orig len doing nothing: 277
   -- after time_added -> t: 232
   -- after att/vic and type->id: 
   Msg("Report len: " .. #s .. "\n")

--   Msg("Roundreport message will be of length: " ..#s .. "\n") 

   -- divide into happy lil bits
   local cut = {}
   while #s != 0 do
      local bit = string.sub(s, 1, 127)
      table.insert(cut, bit)

      s = string.sub(s, 128, -1)
   end
   
   local parts = #cut
--   Msg("Sending msg in parts: " .. parts.. "\n")
   for k, bit in pairs(cut) do
      
      umsg.Start("report_stream")
      umsg.Bool( (k != parts) ) -- continuation bit, true if there's more coming
      umsg.String(bit)
      umsg.End()
   end
   
--   Msg("Message fully sent\n")
end
